#include "config.h"
#include "swai_registry.h"
#include "core/window-private.h"


static SWAIRegistry *swai_registry = NULL;

#define SWAI_REGISTRY_DBUS_SERVICE "org.gnome.Mutter.SwaiRegistry"
#define SWAI_REGISTRY_DBUS_PATH "/org/gnome/Mutter/SwaiRegistry"

static const gchar introspection_xml[] =
  "<node>"
  "  <interface name=\"org.gnome.Mutter.SwaiRegistry\">"
  "    <method name=\"RegisterTitle\">"
  "      <arg name=\"internal_title\" type=\"s\" direction=\"in\"/>"
  "      <arg name=\"display_title\" type=\"s\" direction=\"in\"/>"
  "    </method>"
  "    <signal name=\"title_changed\">"
  "      <arg name=\"internal_title\" type=\"s\" direction=\"out\"/>"
  "      <arg name=\"display_title\" type=\"s\" direction=\"out\"/>"
  "    </signal>"
  "  </interface>"
  "</node>";

static GDBusNodeInfo *introspection_data = NULL;
static guint name_owner_id = 0;

/* Forward declarations */
static void on_bus_acquired(GDBusConnection *connection, const gchar *name, gpointer user_data);
static void on_name_acquired(GDBusConnection *connection, const gchar *name, gpointer user_data);
static void on_name_lost(GDBusConnection *connection, const gchar *name, gpointer user_data);
static void on_window_created(MetaDisplay *display, MetaWindow *window, gpointer user_data);
static void on_window_unmanaging(MetaWindow *window, gpointer user_data);

/* Helper function to check if a window is a SWAI window */
static gboolean is_swai_window(MetaWindow *window)
{
    if (!window || !window->res_class)
        return FALSE;
    
    return g_str_has_prefix(window->res_class, "io.stillhq.swai");
}

/* Helper function to safely add window to list */
static void
swai_registry_add_window(MetaWindow *window)
{
    if (!swai_registry)
        return;
        
    g_mutex_lock(&swai_registry->window_list_mutex);
    
    /* Check if window is already in the list */
    if (g_list_find(swai_registry->window_list, window) == NULL) {
        swai_registry->window_list = g_list_prepend(swai_registry->window_list, window);
        g_debug("SWAI Registry: Added window %s (res_class: %s) to registry", 
                window->desc, window->res_class ? window->res_class : "NULL");
    }
    
    g_mutex_unlock(&swai_registry->window_list_mutex);
}

/* Helper function to safely remove window from list */
static void
swai_registry_remove_window(MetaWindow *window)
{
    if (!swai_registry)
        return;
        
    g_mutex_lock(&swai_registry->window_list_mutex);
    
    if (g_list_find(swai_registry->window_list, window) != NULL) {
        swai_registry->window_list = g_list_remove(swai_registry->window_list, window);
        g_debug("SWAI Registry: Removed window %s from registry", window->desc);
    }
    
    g_mutex_unlock(&swai_registry->window_list_mutex);
}

/* Helper function to update window properties safely */
static void
update_window_properties(MetaWindow *window, const char *display_title, const char *wm_class)
{
    if (!window)
        return;
    
    g_debug("SWAI Registry: Updating window %s - title: '%s' -> '%s', wm_class: '%s' -> '%s'",
            window->desc,
            window->title ? window->title : "NULL", 
            display_title ? display_title : "NULL",
            window->res_class ? window->res_class : "NULL",
            wm_class ? wm_class : "NULL");
    
    /* Update the window title */
    if (display_title && g_strcmp0(window->title, display_title) != 0) {
        meta_window_set_title(window, display_title);
    }
    
    /* Update the window res_class */
    if (wm_class && g_strcmp0(window->res_class, wm_class) != 0) {
        meta_window_set_wm_class(window, wm_class, window->res_name);
    }
}

static char *
swai_registry_remove_window_id(const char *input)
{
    if (!input)
        return NULL;

    char *result = g_strdup(input);
    char *last_dot = g_strrstr(result, ".");

    if (!last_dot)
        return result;

    /* Check if everything after the last dot is numeric */
    char *after_dot = last_dot + 1;
    if (*after_dot == '\0')
        return result; /* No characters after dot */

    gboolean is_numeric = TRUE;
    for (char *p = after_dot; *p != '\0'; p++) {
        if (!g_ascii_isdigit(*p)) {
            is_numeric = FALSE;
            break;
        }
    }

    if (is_numeric) {
        /* Remove the dot and everything after it */
        *last_dot = '\0';
    }

    return result;
}

static void
swai_registry_method_call(
    GDBusConnection *connection,
    const gchar *sender,
    const gchar *object_path,
    const gchar *interface_name,
    const gchar *method_name,
    GVariant *parameters,
    GDBusMethodInvocation *invocation,
    gpointer user_data)
{
    if (g_strcmp0(method_name, "RegisterTitle") == 0) {
        char *internal_title, *display_title;
        g_variant_get(parameters, "(ss)", &internal_title, &display_title);
        
        char *stripped_wm_class = swai_registry_remove_window_id(internal_title);
        
        g_debug("SWAI Registry: RegisterTitle called - internal: '%s', display: '%s', stripped: '%s'",
                internal_title, display_title, stripped_wm_class);
        
        /* Iterate through all windows in the linked list */
        if (swai_registry) {
            g_mutex_lock(&swai_registry->window_list_mutex);
            
            GList *current = swai_registry->window_list;
            int updated_count = 0;
            
            while (current != NULL) {
                MetaWindow *window = (MetaWindow*)current->data;
                current = current->next; /* Move to next before potential list modification */
                
                /* Validate window pointer is still valid */
                if (!window) {
                    continue;
                }
                
                /* Check if this window has res_class beginning with "io.stillhq.swai" */
                if (is_swai_window(window)) {
                    /* Strip numbers from the current window title to compare */
                    char *current_title_stripped = swai_registry_remove_window_id(window->title ? window->title : "");
                    
                    /* If the stripped current title matches the target stripped wm_class */
                    if (g_strcmp0(internal_title, window->title) == 0) {
                        update_window_properties(window, display_title, stripped_wm_class);
                        updated_count++;
                    }
                    
                    g_free(current_title_stripped);
                }
            }
            
            g_mutex_unlock(&swai_registry->window_list_mutex);
            
            g_debug("SWAI Registry: Updated window(s): %s", internal_title);
        }
        
        /* Emit signal */
        if (swai_registry && swai_registry->connection) {
            g_dbus_connection_emit_signal(
                swai_registry->connection,
                NULL,
                SWAI_REGISTRY_DBUS_PATH,
                "org.gnome.Mutter.SwaiRegistry",
                "title_changed",
                g_variant_new("(ss)", internal_title, display_title),
                NULL);
        }
        
        g_dbus_method_invocation_return_value(invocation, NULL);
        
        g_free(internal_title);
        g_free(display_title);
        g_free(stripped_wm_class);
    } else {
        g_dbus_method_invocation_return_error(invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, 
                                            "Unknown method: %s", method_name);
    }
}

static GDBusInterfaceVTable interface_vtable = {
    .method_call = swai_registry_method_call,
    .get_property = NULL,
    .set_property = NULL
};

static void
on_bus_acquired(GDBusConnection *connection, const gchar *name, gpointer user_data)
{
    guint registration_id;
    registration_id = g_dbus_connection_register_object(
        connection,
        SWAI_REGISTRY_DBUS_PATH,
        introspection_data->interfaces[0],
        &interface_vtable,
        NULL, /* user_data */
        NULL, /* user_data_free_func */
        NULL  /* GError** */
    );
    
    if (registration_id == 0) {
        g_warning("Failed to register SwaiRegistry D-Bus object");
    } else {
        swai_registry->connection = g_object_ref(connection);
        swai_registry->registration_id = registration_id;
        g_debug("SWAI Registry: D-Bus object registered successfully");
    }
}

static void
on_name_acquired(GDBusConnection *connection, const gchar *name, gpointer user_data)
{
    g_debug("SWAI Registry: D-Bus name acquired: %s", name);
}

static void
on_name_lost(GDBusConnection *connection, const gchar *name, gpointer user_data)
{
    g_warning("SWAI Registry: D-Bus name lost: %s", name);
}

/* Signal handler for when a window is created */
static void
on_window_created(MetaDisplay *display, MetaWindow *window, gpointer user_data)
{
    if (!window) {
        return;
    }
    
    g_debug("SWAI Registry: Window created: %s (res_class: %s)", 
            window->desc, window->res_class ? window->res_class : "NULL");
    
    /* Add all windows to our list, not just SWAI windows, 
     * so we can track when SWAI windows change their res_class */
    swai_registry_add_window(window);
    
    /* Connect to the unmanaging signal for this window */
    g_signal_connect(window, "unmanaging", G_CALLBACK(on_window_unmanaging), NULL);
}

/* Signal handler for when a window is being destroyed */
static void
on_window_unmanaging(MetaWindow *window, gpointer user_data)
{
    if (!window) {
        return;
    }
    
    g_debug("SWAI Registry: Window unmanaging: %s", window->desc);
    
    /* Remove window from our list */
    swai_registry_remove_window(window);
}

void
swai_registry_init(MetaDisplay *display)
{
    if (swai_registry) {
        g_warning("SWAI Registry: Already initialized");
        return;
    }
    
    g_debug("SWAI Registry: Initializing");
    
    /* Create the registry */
    swai_registry = g_new0(SWAIRegistry, 1);
    swai_registry->window_list = NULL;
    g_mutex_init(&swai_registry->window_list_mutex);
    
    /* Connect to display signals */
    g_signal_connect(display, "window-created", G_CALLBACK(on_window_created), NULL);
    
    /* Initialize D-Bus */
    introspection_data = g_dbus_node_info_new_for_xml(introspection_xml, NULL);
    name_owner_id = g_bus_own_name(
        G_BUS_TYPE_SESSION,
        SWAI_REGISTRY_DBUS_SERVICE,
        G_BUS_NAME_OWNER_FLAGS_NONE,
        on_bus_acquired,
        on_name_acquired,
        on_name_lost,
        NULL,
        NULL
    );
    
    g_debug("SWAI Registry: Initialization complete");
}

void
swai_registry_shutdown(void)
{
    if (!swai_registry) {
        return;
    }
    
    g_debug("SWAI Registry: Shutting down");
    
    /* Clean up D-Bus */
    if (swai_registry->connection && swai_registry->registration_id) {
        g_dbus_connection_unregister_object(swai_registry->connection, swai_registry->registration_id);
    }
    
    if (swai_registry->connection) {
        g_object_unref(swai_registry->connection);
    }
    
    if (introspection_data) {
        g_dbus_node_info_unref(introspection_data);
        introspection_data = NULL;
    }
    
    if (name_owner_id) {
        g_bus_unown_name(name_owner_id);
        name_owner_id = 0;
    }
    
    /* Clean up window list */
    g_mutex_clear(&swai_registry->window_list_mutex);
    g_list_free(swai_registry->window_list);
    
    /* Free the registry */
    g_free(swai_registry);
    swai_registry = NULL;
    
    g_debug("SWAI Registry: Shutdown complete");
}